from . import _linalg as _linalg
from . import _umath_linalg as _umath_linalg
from . import linalg as linalg
from ._linalg import (
    cholesky,
    cond,
    cross,
    det,
    diagonal,
    eig,
    eigh,
    eigvals,
    eigvalsh,
    inv,
    lstsq,
    matmul,
    matrix_norm,
    matrix_power,
    matrix_rank,
    matrix_transpose,
    multi_dot,
    norm,
    outer,
    pinv,
    qr,
    slogdet,
    solve,
    svd,
    svdvals,
    tensordot,
    tensorinv,
    tensorsolve,
    trace,
    vecdot,
    vector_norm,
)

__all__ = [
    "LinAlgError",
    "cholesky",
    "cond",
    "cross",
    "det",
    "diagonal",
    "eig",
    "eigh",
    "eigvals",
    "eigvalsh",
    "inv",
    "lstsq",
    "matmul",
    "matrix_norm",
    "matrix_power",
    "matrix_rank",
    "matrix_transpose",
    "multi_dot",
    "norm",
    "outer",
    "pinv",
    "qr",
    "slogdet",
    "solve",
    "svd",
    "svdvals",
    "tensordot",
    "tensorinv",
    "tensorsolve",
    "trace",
    "vecdot",
    "vector_norm",
]

class LinAlgError(ValueError): ...
