/*
 * This file exports the private function that exposes the DType API
 *
 * This file is a stub, all important definitions are in the code file.
 */

#ifndef NUMPY_CORE_SRC_MULTIARRAY_EXPERIMENTAL_PUBLIC_DTYPE_API_H_
#define NUMPY_CORE_SRC_MULTIARRAY_EXPERIMENTAL_PUBLIC_DTYPE_API_H_

NPY_NO_EXPORT void
_fill_dtype_api(void *numpy_api_table[]);


#endif  /* NUMPY_CORE_SRC_MULTIARRAY_EXPERIMENTAL_PUBLIC_DTYPE_API_H_ */
