#include <assert.h>
#include <stdlib.h>

int meson_test_side_effect = EXIT_FAILURE;

int meson_test_set_side_effect(void) {
    meson_test_side_effect = EXIT_SUCCESS;
    return 1;
}

int main(void) {
    // meson_test_side_effect is set only if assert is executed
    assert(meson_test_set_side_effect());
    return meson_test_side_effect;
}
