#ifndef _NPY_CORE_SRC_UMATH_STRINGDTYPE_UFUNCS_H_
#define _NPY_CORE_SRC_UMATH_STRINGDTYPE_UFUNCS_H_

#ifdef __cplusplus
extern "C" {
#endif

NPY_NO_EXPORT int
init_stringdtype_ufuncs(PyObject* umath);

#ifdef __cplusplus
}
#endif

#endif /* _NPY_CORE_SRC_UMATH_STRINGDTYPE_UFUNCS_H_ */
